/*
 * Ambient - the ultimate desktop
 * ------------------------------
 * Copyright 2001-2005 by David Gerber <zapek@morphos.net>
 * Copyright 2006-2012 Ambient Open Source Team
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Id: threads_tp.c,v 1.2 2013/02/12 20:26:24 itix Exp $
 */

#include "ambient.h"

/* public */
#include <exec/execbase.h>
#include <clib/alib_protos.h>
#include <proto/exec.h>
#include <dos/dostags.h>
#include <dos/dosextens.h>
#include <libraries/threadpool.h>
#include <proto/dos.h>
#include <utility/tagitem.h>
#include <proto/threadpool.h>
#include <proto/utility.h>

/* private */
#include "dosreq.h"
#include "print.h"
#include "prefs_desktop.h"
#include "threads.h"
#include "methodstack.h"
#include "mui_func.h"
#include "prefs_startup.h"
#include "scandir.h"
#include "background.h"
#include "threadtags.h"
#include "wbstartup.h"
#include "devices.h"
#include "iconio.h"
#include "taskdata.h"
#include "getdirsize.h"
#include "findver.h"
#include "movedir.h"
#include "movelist.h"
#include "deleteall.h"
#include "format.h"
#include "wbstart.h"
#include "ipc.h"
#include "snapshot.h"
#include "snapshotlist.h"
#include "rexx.h"
#include "md5sum.h"
#include "appmessage.h"
#include "prefsio.h"
#include "panelprefs.h"
#include "paneleffects.h"
#include "appicon.h"
#include "infowin.h"
#include "shortcuts.h"
#include "loaduri.h"
#include "rename.h"
#include "tags.h"
#include "playsound.h"
#include "loadimage.h"
#include "imageview.h"
#include "textview.h"
#include "boopsiview.h"
#include "makedir.h"
#include "makelink.h"
#include "fonts.h"
#include "infoicongroup.h"
#include "thumbs.h"
#include "smartreq.h"
#include "metadata.h"
#include "actiondispatcherclass.h"
#include "typescanner.h"
#include "clipboard.h"
#include "fastcopy.h"
#include "dragdrop.h"
#include "listviewclass.h"
#include "debug.h"
#include "findclass.h"
#include "appclass.h"
#include "paneltags.h"

#if USE_THREADPOOL
#define THREAD_MIN 2                   /* minimum number of threads in the pool */
#define THREAD_MAX 9999                /* maximum, unlikely to be reached any day :) */
#define THREAD_STACK STACKSIZE_THREAD  /* stack per thread */
#define THREAD_WAITCLEANUP 5           /* time in seconds to wait for the thread cleanup on exit */
#define THREAD_MAXABORT 8UL            /* maximum number of aborting tags */

#define THREAD_PRIORITY 0

/*
 * Define if you want to have info on the action ID the thread performa.
 */

#define THREAD_ACTION_INFO_HACK

struct thread_msg {
	struct MinNode n;
	struct Process *proc;
	APTR destobj;
	ULONG state;
	LONG pri;
	size_t work_id;
	ULONG action;
	ULONG status;
	APTR obj;
	struct TagItem *taglist;
};

static struct SignalSemaphore semaphore;
static struct MinList threadlist;
static struct MinList finishlist;

static BYTE threadsigbit = -1;
static APTR threadpool;

static void ambient_thread(struct thread_msg *msg, struct MsgPort *port);
static ULONG thread_domsg(struct thread_msg *msg);

ULONG threadsig;

struct Task *maintask;

/*
 * Formats thread name according to action ID.
 */

#define PROCESS_THREAD_NAME "Ambient Thread "
#define PROCESS_THREAD_ACTION_LEN 10
#define PROCESS_THREAD_ACTION_DEF "     "

static void thread_build_name( STRPTR buf, ULONG bufsize, LONG action )
{
	switch (action)
	{
		case TA_Devices_UpdateInfo:
			snprintf( buf, bufsize, "%s [%lu] (%s)", PROCESS_THREAD_NAME, action, "UpdateInfo" );
			break;

		default:
			snprintf( buf, bufsize, "%s [%lu]", PROCESS_THREAD_NAME, action );
			break;
	}
}

/*
 * Initializes the whole thread system.
 */
ULONG threads_init(void)
{
	InitSemaphore(&semaphore);

	NEWLIST(&threadlist);
	NEWLIST(&finishlist);

	D(PROC, bug("creating threads support..\n"));

	maintask = FindTask(NULL);

	if ((threadsigbit = AllocSignal(-1)) >= 0)
	{
		threadsig = 1L << threadsigbit;

		threadpool = CreateThreadPoolTagList(THREAD_MAX, NULL);

		if (threadpool)
			return (TRUE);
	}

	return (FALSE);
}


/*
 * Cleanups the whole thread system.
 */
ULONG threads_finish(ULONG loop)
{
	DB(("Aborting threads..:%d.\n", loop))

	if (threadpool)
	{
		struct thread_msg *msg;

		/*
		 * Tell every process to die
		 */
		D(PROC, bug("telling processes to die..\n"));

		ObtainSemaphoreShared(&semaphore);
		ITERATELIST(msg, &threadlist)
		{
			D(PROC, bug("signaling %p (%d)..\n", msg->proc, msg->action ));
			SignalWorkItem(threadpool, msg->work_id, SIGBREAKF_CTRL_D);
		}
		ReleaseSemaphore(&semaphore);

		D(PROC, bug("all processes told to die, waiting %lu seconds for them to finish\n", (ULONG)THREAD_WAITCLEANUP));

		if (!ISLISTEMPTY(&threadlist))
		{
			if (loop)
			{
				ULONG trynum = THREAD_WAITCLEANUP * 5 + 1;

				while (!ISLISTEMPTY(&threadlist))
				{
					D(PROC, bug("processes still alive.. returning to try again\n"));
					Delay(10);

					if (trynum)
					{
						trynum--;

						if (!trynum)
						{
							/*
							 * This case never happens actually.
							 * threads_finish() is called from mainloop and
							 * relies on mui ticks.
							 */
							smartreq_info("Ambient Threads", MV_Notification_Warning, "%lu remaining thread(s) are preventing the cleanup sequence.\nWaiting more..", trynum);
						}
					}
				}
			}
			else
			{
				return (FALSE);
			}
		}

		D(PROC,bug("all processed dead\n"));
		threads_handle();

		D(PROC,bug("remove threadpool\n"));
		DeleteThreadPool(threadpool);
		threadpool = NULL;
		D(PROC, bug("threadpool removed\n"));
	}

	FreeSignal(threadsigbit);

	threadsigbit = -1;

	return (TRUE);
}


void threads_cleanup(void)
{
	threads_finish(TRUE);
}


/*
 * Remove the threads in IDLE state.
 */
void threads_reclaim(void)
{
}


/*
 * Signal all the threads dealing with 'obj' (MM_Thread_Finished) to
 * piss off. MM_Thread_Finished
 * with a msg->id of 0 is sent when they are aborted or
 * if there's nothing. msg->status tells if there were aborts.
 * obj - is the TT_Object by default, can be changed using TT_Object in special cases
 */
void threads_abort(APTR obj, ...)
{
	ULONG found = FALSE;
	struct thread_msg *msg;
	va_list va;
	ULONG specific;

	MAINTASK;

	va_start(va, obj);
	specific = va_arg(va, ULONG);
	va_end(va);

	ObtainSemaphoreShared(&semaphore);
	/*
	 * Check if there's something to signal at
	 * all.
	 */
	ITERATELIST(msg, &threadlist)
	{
		if (msg->destobj == obj && msg->state == TS_BUSY)
		{
			if (specific)
			{
				ULONG t;

				va_start(va, obj);
				while ( (t = va_arg(va, ULONG)) )
				{
					if (t == msg->action)
					{
						D(PROC,bug("signaling process %p (action %d) to stop..\n", msg->proc, msg->action));

						SignalWorkItem(threadpool, msg->work_id, SIGBREAKF_CTRL_D);
						found = TRUE;
						break;
					}
				}
				va_end(va);
			}
			else
			{
				D(PROC, bug("signaling process %p (action %d) to stop..\n", msg->proc, msg->action));

				SignalWorkItem(threadpool, msg->work_id, SIGBREAKF_CTRL_D);
				found = TRUE;
			}
		}
	}
	ReleaseSemaphore(&semaphore);

	if (!found)
	{
		D(PROC, bug("no thread found, sending finished method to obj %p directly..\n", obj));
		/*
		 * Following is always sent when there's nothing.
		 */
		if (specific)
		{
			ULONG t;

			va_start(va, obj);
			while ( (t = va_arg(va, ULONG)) )
			{
				methodstack_push_sync(obj, 4, MM_Thread_Finished, t, MV_Thread_Finished_Abort, NULL);
			}
			va_end(va);
		}
		else
		{
			methodstack_push_sync(obj, 4, MM_Thread_Finished, NULL, MV_Thread_Finished_Abort, NULL);
		}
	}
}

/*
 * Puts current (running) thread to sleep.
 */

void thread_wait( void )
{
	struct thread_msg *msg;
	struct Process *me = (APTR)SysBase->ThisTask;

	THREAD;

	ObtainSemaphoreShared(&semaphore);

	ITERATELIST(msg, &threadlist)
	{
		if ( msg->proc == me )
		{
			msg->state = TS_WAITING;
			ReleaseSemaphore(&semaphore);
			Wait( SIGBREAKF_CTRL_D );
			msg->state = TS_BUSY;
			return;
		}
	}

	ReleaseSemaphore(&semaphore);
}

APTR thread_get( void )
{
	THREAD;

	if ( IS_MAINTASK )
		return NULL;
	else
		return FindTask( NULL );

}

/*
 * Try to awake sleeping thread. If the thread is not
 * in 'sleeping' state then the function does nothing.
 */

void thread_signal(APTR thread)
{
	struct thread_msg *msg;

	MAINTASK;

	ObtainSemaphoreShared(&semaphore);

	/*
	 * Check if there's something to signal at
	 * all.
	 */

	ITERATELIST(msg, &threadlist)
	{
		if (msg->proc == thread && msg->state == TS_WAITING)
		{
			D(PROC,bug("signaling process %p..\n", msg->proc));
			SignalWorkItem(threadpool, msg->work_id, SIGBREAKF_CTRL_D);
			break;
		}
	}

	ReleaseSemaphore(&semaphore);
}

/*
 * Do an action. Spawns a thread if necessary.
 */
ULONG v_do_action(APTR obj, int action, struct TagItem *tags)
{
	struct thread_msg *msg;

	MAINTASK;

	if ( (msg = malloc(sizeof(*msg))) )
	{
		struct TagItem *taglist;

		if ( (taglist = threadtags_clone(tags)) )
		{
			va_end(va);

			D(PROC, bug("sending message to thread (direct mode, action: %ld)\n", (LONG)action));

			msg->proc = NULL;
			msg->state = TS_BUSY;
			msg->obj              = obj;
			msg->taglist          = taglist;

			msg->pri = GetTagData(TT_Priority, 0, msg->taglist);
			msg->destobj = (APTR)GetTagData(TT_Object, (ULONG)obj, msg->taglist);
			msg->action  = action;

			ObtainSemaphore(&semaphore);
			ADDTAIL(&threadlist, msg);
			ReleaseSemaphore(&semaphore);

			msg->work_id = QueueWorkItem(threadpool, (THREADFUNC)ambient_thread, msg);

			return (TRUE); /* XXX ? */
		}
		free(msg);
	}

	/* XXX: hum.. */
	return (FALSE);
}

/*
 * Do an action. It's NOT spawning new thread, but have to be executed from
 * already existing one. It will return action code.
 */
ULONG v_do_action_sync(APTR obj, int action, struct TagItem *taglist)
{
	struct thread_msg msg;
	APTR destobj;
	ULONG status;

	THREAD;

	DB(("Executing synchronous action:%d\n", action ));

	destobj = (APTR)GetTagData(TT_Object, (ULONG)obj, taglist);

	msg.action           = action; /* thread_domsg() will use this one when msg->tn == NULL */
	msg.status           = 0;      /* not really needed */
	msg.obj              = obj;
	msg.taglist          = taglist;

	status               = thread_domsg(&msg);

	DB(("Action returned (%d)!\n", status ));

	D(PROC, bug("sending MM_Thread_Finished\n"));
	if ( destobj )
		methodstack_push_sync(destobj, 4, MM_Thread_Finished, action, status, taglist);

	return status;
}


/*
 * Handles the thread system after receiving a replymsg.
 */
void threads_handle(void)
{
	struct thread_msg *msg;

	ObtainSemaphore(&semaphore);

	ITERATELIST(msg, &finishlist)
	{
		if (msg->destobj)
		{
			D(PROC, bug("sending MM_Thread_Finished (action %d)\n", msg->action));
			methodstack_push_sync(msg->destobj, 4, MM_Thread_Finished, msg->action, msg->status, msg->taglist);
		}

#if 0
				D(PROC, bug("not found, sending method to %p..\n", sn->obj));
				if (sn->abort[0])
				{
					ULONG i;

					for (i = 0; i < THREAD_MAXABORT; i++)
					{
						if (sn->abort[i])
						{
							methodstack_push_sync(sn->obj, 4, MM_Thread_Finished, sn->abort[i], MV_Thread_Finished_Abort, msg->taglist);
						}
					}
				}
				else
				{
					methodstack_push_sync(sn->obj, 4, MM_Thread_Finished, NULL, MV_Thread_Finished_Abort, msg->taglist);
				}
#endif

		D(PROC,bug("Releasing thread args...\n"));
		threadtags_free(msg->taglist);
		D(PROC,bug("Done\n"));
		free(msg);
	}

	NEWLIST(&finishlist);

	ReleaseSemaphore(&semaphore);
}


ULONG threads_check_abort(void)
{
	return (SetSignal(0, SIGBREAKF_CTRL_D) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D));
}


ULONG threads_waitsig(ULONG usersigs, ULONG *aborted)
{
	ULONG sigmask = Wait(SIGBREAKF_CTRL_D | SIGBREAKF_CTRL_C | usersigs);

	*aborted = sigmask & (SIGBREAKF_CTRL_D | SIGBREAKF_CTRL_C);

	return sigmask;
}


/*
 * Actual threads.
 */
static void ambient_thread(struct thread_msg *msg, struct MsgPort *port)
{
	struct Process *me = (APTR)SysBase->ThisTask;
	ULONG status = ABORTED;

	msg->proc = me;

	if (!threads_check_abort())
	{
		struct taskdata td;
		TEXT buf[80];

		if (me->pr_Task.tc_Node.ln_Pri != msg->pri)
			SetTaskPri(&me->pr_Task, msg->pri);

		dosreq_disable();
		thread_build_name(buf, sizeof(buf), msg->action);

		td.msport = port;
		td.reqport = NULL;

		me->pr_Task.tc_Node.ln_Name = buf;
		me->pr_Task.tc_UserData = &td;

		D(PROC, bug("%p working..\n", me));
		status = thread_domsg(msg);
	}

	msg->status = status;

	#ifdef DEBUG
	if (status == (ULONG)ABORTED)
	{
		D(PROC, bug("%p marking ABORTED\n", me));
	}
	#endif

	D(PROC, bug("%p did its job number %ld\n", me, msg->action));

	ObtainSemaphore(&semaphore);
	REMOVE(msg);
	ADDTAIL(&finishlist, msg);
	ReleaseSemaphore(&semaphore);

	/* Release first, signal after. */
	Signal(maintask, threadsig);
}


static ULONG thread_domsg(struct thread_msg *msg)
{
	ULONG action = msg->action;

	D(PROC, bug("processing message action %ld..\n", action));

	switch (action)
	{
		case TA_Slack:
			return (TRUE);

		case TA_File_ScanDir:
			return (tr_scandir(msg->obj,
				(STRPTR)GetTagData(TT_File_ScanDir_Path, NULL, msg->taglist),
				GetTagData(TT_File_ScanDir_Mode, TV_File_ScanDir_Mode_Files, msg->taglist)
			));

		case TA_Background_Load:
			return (tr_background_load(msg->obj,
				GetTagData(TT_Background_Load_Type, 0, msg->taglist),
				GetTagData(TT_Background_Load_Mode, 0, msg->taglist),
				(STRPTR)GetTagData(TT_Background_Load_Path, NULL, msg->taglist)
			));

		case TA_WBStartup_Execute:
			return (tr_wbstartup_execute((STRPTR)GetTagData(TT_WBStartup_Execute_Path, NULL, msg->taglist)
			));

		case TA_Devices_Show:
			return (tr_devices_show(msg->obj,
				GetTagData(TT_Devices_Show_Assigns, FALSE, msg->taglist),
				GetTagData(TT_Devices_Show_IsRoot, FALSE, msg->taglist),
				FALSE
			));

		case TA_File_Delete:
			return (tr_deleteall(msg->obj,
				(APTR)GetTagData(TT_File_Delete_Refwin, NULL, msg->taglist),
				(STRPTR *)GetTagData(TT_File_Delete_PathList, NULL, msg->taglist),
				(ULONG)GetTagData(TT_File_Delete_NoIcon, NULL, msg->taglist)
			));

		case TA_Icon_Snapshot:
			{
				struct MinList *l = (struct MinList *)GetTagData(TT_Icon_Snapshot_List, NULL, msg->taglist);

				if (l)
				{
					/* NOTE: tr_snapshot_list frees the nodes and the list! */
					return (tr_snapshot_list(l));
				}
				else
				{
					return (tr_snapshot_icon(
						(STRPTR)GetTagData(TT_Icon_Snapshot_Path, NULL, msg->taglist),
						(LONG)GetTagData(TT_Icon_Snapshot_X, 0, msg->taglist),
						(LONG)GetTagData(TT_Icon_Snapshot_Y, 0, msg->taglist)
					));
				}
			}
			break;

		case TA_Icon_Unsnapshot:
			return (tr_snapshot_icon(
				(STRPTR)GetTagData(TT_Icon_Snapshot_Path, NULL, msg->taglist),
				NO_ICON_POSITION, NO_ICON_POSITION
			));
			break;

		case TA_Window_Snapshot:
			return (tr_snapshot_window(
				(STRPTR)GetTagData(TT_Window_Snapshot_Path, NULL, msg->taglist),
				(LONG)GetTagData(TT_Window_Snapshot_X, 0, msg->taglist),
				(LONG)GetTagData(TT_Window_Snapshot_Y, 0, msg->taglist),
				GetTagData(TT_Window_Snapshot_XS, 0, msg->taglist),
				GetTagData(TT_Window_Snapshot_YS, 0, msg->taglist),
				GetTagData(TT_Window_Snapshot_Flags, 0, msg->taglist)
			));

		case TA_File_GetSize:
			return (tr_getdirsize(msg->obj,
				(STRPTR)GetTagData(TT_File_GetSize_Path, NULL, msg->taglist)
			));

		case TA_File_FindVer:
			return (tr_findver(msg->obj,
				(STRPTR)GetTagData(TT_File_FindVer_Path, NULL, msg->taglist)
			));

		case TA_File_Move:
			return(tr_move(msg->obj,
					(APTR)GetTagData(TT_File_Move_Refwin, NULL, msg->taglist),
					(struct MinList *)GetTagData(TT_File_Move_SrcList, NULL, msg->taglist),
					(STRPTR)GetTagData(TT_File_Move_SrcPath, NULL, msg->taglist),
					(STRPTR)GetTagData(TT_File_Move_DstPath, NULL, msg->taglist),
					GetTagData(TT_File_Move_Copy, FALSE, msg->taglist),
					GetTagData(TT_File_Move_NoIcon, FALSE, msg->taglist),
					GetTagData(TT_File_Move_Rename, FALSE, msg->taglist)
			));

		case TA_AppMsg_Send:
			return (tr_appmessage_send(
				GetTagData(TT_AppMsg_Send_Type, NULL, msg->taglist),
				(struct MsgPort *)GetTagData(TT_AppMsg_Send_MsgPort, NULL, msg->taglist),
				(struct Window *)GetTagData(TT_AppMsg_Send_Window, NULL, msg->taglist),
				(CONST_STRPTR)GetTagData(TT_AppMsg_Send_Path, NULL, msg->taglist),
				GetTagData(TT_AppMsg_Send_Class, NULL, msg->taglist),
				GetTagData(TT_AppMsg_Send_ID, NULL, msg->taglist),
				GetTagData(TT_AppMsg_Send_Userdata, NULL, msg->taglist),
				GetTagData(TT_AppMsg_Send_NumArgs, 0, msg->taglist),
				(struct WBArg *)GetTagData(TT_AppMsg_Send_WBArgList, NULL, msg->taglist),
				GetTagData(TT_AppMsg_Send_MouseX, 0, msg->taglist),
				GetTagData(TT_AppMsg_Send_MouseY, 0, msg->taglist)
			));

		case TA_Devices_Add:
			return (tr_devices_add(msg->obj,
				GetTagData(TT_Devices_Fade, TRUE, msg->taglist)
			));

		case TA_Devices_Remove:
			return (tr_devices_remove(msg->obj));

		case TA_Devices_RemoveAll:
			return (tr_devices_removeall(msg->obj));

		case TA_Devices_UpdateInfo:
			return (tr_devices_updateinfo(msg->obj));

		case TA_Disk_Format:
			return (tr_format(msg->obj,
				(STRPTR)GetTagData(TT_Disk_Format_Device, NULL, msg->taglist),
				(struct device_info *)GetTagData(TT_Disk_Format_DeviceInfo, NULL, msg->taglist),
				GetTagData(TT_Disk_Format_Mode, NULL, msg->taglist),
				GetTagData(TT_Disk_Format_FileSystem, NULL, msg->taglist),
				GetTagData(TT_Disk_Format_Flags, 0, msg->taglist)
			));

		case TA_Icon_Write:
			{
				ULONG flags = GetTagData(TT_Icon_Write_Flags, 0, msg->taglist);
				STRPTR comment = (STRPTR)GetTagData(TT_Icon_Write_Comment, NULL, msg->taglist);

				if (comment || flags)
				{
					return (tr_set_file_and_write_icon(msg->obj,
						(STRPTR)GetTagData(TT_Icon_Write_Path, NULL, msg->taglist),
						flags,
						comment
					));
				}
				else
				{
					return (tr_icon_write(msg->obj,
						(STRPTR)GetTagData(TT_Icon_Write_Path, NULL, msg->taglist)
					));
				}
			}
			break;

		case TA_Icon_GetProperties:
			{
				struct ipcmessage *imsg = (struct ipcmessage *)GetTagData(TT_Icon_GetProperties_Message, NULL, msg->taglist);
				struct ipc_wbstartlib *wbsm = imsg->msgtype;
				ASSERT(imsg);
				ASSERT(wbsm);

				wbsm->status = get_icon_properties(wbsm->filename, wbsm->type, wbsm->pri, wbsm->stacksize, wbsm->defaulttool);
				ReplyMsg(&imsg->msg);
			}
			return (TRUE);

		case TA_File_MD5sum:
			{
				struct MinList *l = (struct MinList *)GetTagData(TT_File_MD5sum_List, NULL, msg->taglist);

				if (l)
				{
					return (tr_md5sum_list(msg->obj,
						l
					));
				}
				else
				{
					return (tr_md5sum(msg->obj,
						(STRPTR)GetTagData(TT_File_MD5sum_Path, NULL, msg->taglist)
					));
				}
			}
			break;

		case TA_Prefs_Save:
			return (tr_prefs_save(msg->obj,
				(APTR)GetTagData(TT_Prefs_Save_Ctx, NULL, msg->taglist)
			));

		case TA_Panels_LoadAll:
			return (tr_panels_loadall());

		case TA_Panels_Save:
			return (tr_panels_save(msg->obj,
				(APTR)GetTagData(TT_Panels_Save_Ctx, NULL, msg->taglist)
			));

		case TA_Panels_Delete:
			return (tr_panels_delete(msg->obj,
				(APTR)GetTagData(TT_Panels_Delete_Ctx, NULL, msg->taglist)
			));

		case TA_Panels_Move:
			return (tr_panels_move(msg->obj,
				(struct Window *)GetTagData(TT_Panels_Move_Window, NULL, msg->taglist),
				GetTagData(TT_Panels_Move_X, 0, msg->taglist),
				GetTagData(TT_Panels_Move_Y, 0, msg->taglist)
			));

		case TA_Panels_Zip:
			return (tr_panels_zip(msg->obj,
				(struct Window *)GetTagData(TT_Panels_Zip_Window, NULL, msg->taglist),
				GetTagData(TT_Panels_Zip_XS, 0, msg->taglist),
				GetTagData(TT_Panels_Zip_YS, 0, msg->taglist),
				GetTagData(TT_Panels_Zip_Reversed, FALSE, msg->taglist),
				GetTagData(TT_Panels_Zip_ZipSpeed, MV_Panel_ZipSpeed_Medium, msg->taglist)
			));

		case TA_Appicon_Read:
			{
				ULONG rc;

				struct ipcmessage *imsg = (struct ipcmessage *)GetTagData(TT_Appicon_Read_Message, NULL, msg->taglist);
				ASSERT(imsg);

				rc = tr_appicon_read(msg->obj, (APTR)GetTagData(TT_Appicon_Read_Object, NULL, msg->taglist), (struct ipc_appicon *)imsg->msgtype);
				return (rc);
			}

		case TA_Infowin_Open:
			return (tr_openinfowin(msg->obj,
				(CONST CONST_STRPTR *)GetTagData(TT_Infowin_Open_PathList, NULL, msg->taglist),
				GetTagData(TT_Infowin_Open_Wait, FALSE, msg->taglist),
				GetTagData(TT_Infowin_Open_RexxID, NULL, msg->taglist)
			));

		#if USE_SHORTCUTS
		case TA_Shortcuts_Add:
			{
				struct MinList *l = (struct MinList *)GetTagData(TT_Shortcuts_Add_List, NULL, msg->taglist);

				if (l)
				{
					/* NOTE: tr_shotrcuts_list frees the nodes and the list! */
					return (tr_shortcuts_list_add(msg->obj,
						l
					));
				}
				else
				{
					return (tr_shortcuts_add(msg->obj,
						(STRPTR)GetTagData(TT_Shortcuts_Add_Path, NULL, msg->taglist),
						(LONG)GetTagData(TT_Shortcuts_Add_X, NO_ICON_POSITION, msg->taglist),
						(LONG)GetTagData(TT_Shortcuts_Add_Y, NO_ICON_POSITION, msg->taglist),
						TRUE,
						(LONG)GetTagData(TT_Shortcuts_Add_Type, MV_Icon_FileType_None, msg->taglist)
					));
				}
			}
			break;

		case TA_Shortcuts_SaveAll:
			return (tr_shortcuts_save(msg->obj,
				TRUE
			));
		#endif

		case TA_URI_Load:
			return (tr_loaduri(
				(APTR)GetTagData(TT_URI_Load_Ctx, NULL, msg->taglist),
				(STRPTR)GetTagData(TT_URI_Load_URI, NULL, msg->taglist),
				GetTagData(TT_URI_Load_ID, 0, msg->taglist)
			));

		case TA_File_Rename:
			{
				ULONG reqflags = 0;
				return (tr_rename(
					(STRPTR)GetTagData(TT_File_Rename_Path, NULL, msg->taglist),
					(CONST_STRPTR *)GetTagData(TT_File_Rename_PathList, NULL, msg->taglist),
					(STRPTR)GetTagData(TT_File_Rename_Name, NULL, msg->taglist),
					GetTagData(TT_File_Rename_NoIcon, (ULONG) FALSE, msg->taglist),
					TRUE,
					(ULONG *)&reqflags
				));
			}

		case TA_Sound_Play:
			return (tr_playsound(
				(STRPTR)GetTagData(TT_Sound_Play_Path, NULL, msg->taglist),
				GetTagData(TT_Sound_Play_Mode, PS_VORBIS | PSF_QUEUED_IMMEDIATE, msg->taglist)
			));

		case TA_Imageview_Load:
			return (tr_loadimage(msg->obj,
				(STRPTR)GetTagData(TT_Imageview_Load_Path, NULL, msg->taglist)
			));

		case TA_Imageview_Scale:
			return (tr_scaleimage(msg->obj,
				(APTR)GetTagData(TT_Imageview_Scale_Dtp, NULL, msg->taglist),
				GetTagData(TT_Imageview_Scale_XS, NULL, msg->taglist),
				GetTagData(TT_Imageview_Scale_YS, NULL, msg->taglist)
			));

		case TA_Icon_Read:
			return (tr_icon_read(msg->obj,
				(STRPTR)GetTagData(TT_Icon_Read_Path, NULL, msg->taglist),
				GetTagData(TT_Icon_Read_NoInfo, FALSE, msg->taglist)
			));

		case TA_Imageview_ScanPics:
			return (tr_scanforpics(msg->obj,
				(STRPTR)GetTagData(TT_Imageview_ScanPics_Path, NULL, msg->taglist)
			));

		case TA_Textview_Load:
			return (tr_loadtext(msg->obj,
				(STRPTR)GetTagData(TT_Textview_Load_Path, NULL, msg->taglist)
			));

		case TA_Boopsiview_Load:
			return (tr_loaddatatype(msg->obj,
				(STRPTR)GetTagData(TT_Boopsiview_Load_Path, NULL, msg->taglist)
			));

		case TA_File_Makedir:
			return (tr_makedir(
				(STRPTR)GetTagData(TT_File_Makedir_Path, NULL, msg->taglist),
				GetTagData(TT_File_Makedir_Icon, FALSE, msg->taglist)
			));

		case TA_File_Makelink:
			return (tr_makelink(
				(STRPTR)GetTagData(TT_File_Makelink_Object, NULL, msg->taglist),
				(STRPTR)GetTagData(TT_File_Makelink_Link, NULL, msg->taglist)
			));

		case TA_Font_Load:
			DB(("THREAD:LOAD_FONT:%s:%d\n",(STRPTR)GetTagData(TT_Font_Load_Path, NULL, msg->taglist),GetTagData(TT_Font_Load_Type, TV_Font_Load_Type_Root, msg->taglist)));
			return (tr_font_load(msg->obj,
				(STRPTR)GetTagData(TT_Font_Load_Path, NULL, msg->taglist),
				GetTagData(TT_Font_Load_Type, TV_Font_Load_Type_Root, msg->taglist)
			));

		case TA_WBStart:
			return (tr_wbstart(
				(STRPTR)GetTagData(TT_WBStart_Path, NULL, msg->taglist),
				(LONG)GetTagData(TT_WBStart_Priority, 0, msg->taglist),
				GetTagData(TT_WBStart_Stack, 0, msg->taglist),
				(STRPTR)GetTagData(TT_WBStart_Argument, NULL, msg->taglist),
				(struct MinList *)GetTagData(TT_WBStart_Argument_List, NULL, msg->taglist),
				GetTagData(TT_WBStart_FreeNames, FALSE, msg->taglist)
			));

		case TA_Infoicon_Load:
			return (tr_infoicon_load(msg->obj,
				(STRPTR)GetTagData(TT_Infoicon_Load_Path, NULL, msg->taglist)
			));

		case TA_Thumbnail_Create:
			{
				struct MinList *l = (struct MinList *)GetTagData(TT_Thumbnail_Create_List, NULL, msg->taglist);

				if (l)
				{
					return (tr_thumb_createicon_list(msg->obj,
						l,
						TCF_CREATE
					));
				}
				else
				{
					return (tr_thumb_createicon(msg->obj,
						(APTR)GetTagData(TT_Thumbnail_Create_Object, NULL, msg->taglist),
						(STRPTR)GetTagData(TT_Thumbnail_Create_Path, NULL, msg->taglist),
						TCF_CREATE
					));
				}
			}

		case TA_Metadata_Gather:
			return (tr_metadata_gather(msg->obj,
				(STRPTR)GetTagData(TT_Metadata_Gather_Path, NULL, msg->taglist),
				(STRPTR)GetTagData(TT_Metadata_Gather_MimeTypeDescription, NULL, msg->taglist)
			));

		case TA_ActionDispatcher_Execute:
			return (tr_actiondispatcher_execute(msg->obj
			));

		case TA_MimeType_Scan:
			{
				struct MinList *l = (struct MinList *)GetTagData(TT_MimeType_Scan_List, NULL, msg->taglist);
				ULONG matchfirst = GetTagData(TT_MimeType_Scan_MatchFirst, FALSE, msg->taglist);

				if ( matchfirst )
				{
					APTR *mimetypeptr = (APTR*)GetTagData(TT_MimeType_Scan_MimetypePtr, NULL, msg->taglist);
					return (tr_typescanner_scan_matchfirst(msg->obj, l, mimetypeptr));
				}
				else
				{
					if ( l )
					{
						ULONG gen_thumbs = GetTagData(TT_MimeType_Scan_GenerateThumbs, FALSE, msg->taglist);
						ULONG gen_deficons = GetTagData(TT_MimeType_Scan_GenerateDefIcons, FALSE, msg->taglist);
						return (tr_typescanner_scan_list(msg->obj, l, gen_deficons, gen_thumbs));
					}
					else
					{
						return (tr_typescanner_scan_entry(msg->obj,
							(APTR)GetTagData(TT_MimeType_Scan_Path, NULL, msg->taglist),
							(APTR)GetTagData(TT_MimeType_Scan_MimetypePtr, NULL, msg->taglist)
						));
					}
				}
			}

		case TA_Clipboard_Paste:
			return (tr_clipboard_paste(
				(STRPTR) GetTagData(TT_Clipboard_Paste_DestPath , 0, msg->taglist),
				(LONG) GetTagData(TT_Clipboard_Paste_ViewID, 0, msg->taglist),
				(LONG) GetTagData(TT_Clipboard_Paste_As, FALSE, msg->taglist)
			));

		case TA_Listview_CreateIcons:
			return (tr_listview_createicons(msg->obj,
				(APTR) GetTagData(TT_Listview_Iconlist, NULL, msg->taglist),
				(APTR) GetTagData(TT_Listview_Data, NULL, msg->taglist)
			));

		case TA_Listview_ShowDevices:
			return (tr_listview_showdevices(msg->obj,
			    (APTR) GetTagData(TT_Listview_Data, NULL, msg->taglist)
			));

		case TA_Find:
			return (tr_find(msg->obj,
				(APTR) GetTagData(TT_Find_Window, NULL, msg->taglist)
			));

		case TA_DesktopPrefs_Save:
			return (tr_dprefs_save(msg->obj));

		case TA_Version_Find:
			return (tr_findver_list(msg->obj,
				(struct MinList *)GetTagData(TT_Version_Find_List, NULL, msg->taglist)
			));

		case TA_File_GetSizes:
			return (tr_getdirsizes(msg->obj,
				(struct MinList *) GetTagData(TT_File_GetSizes_List, NULL, msg->taglist)
			));

		case TA_Dispose_Objects:
			return (tr_dispose_objects(msg->obj,
				(APTR*)GetTagData(TT_Dispose_Objects_Array, NULL, msg->taglist),
				(ULONG)GetTagData(TT_Dispose_Objects_FreeArray, TRUE, msg->taglist)
			));

		case TA_Listview_ShowPreview:
			return (tr_listview_showpreview(msg->obj,
				(APTR) GetTagData(TT_Listview_Entry, NULL, msg->taglist),
				(APTR) GetTagData(TT_Listview_Data, NULL, msg->taglist)
			));

		case TA_Iconview_ShowPreview:
			return (tr_iconview_showpreview(msg->obj,
				(APTR) GetTagData(TT_Iconview_ShowPreview_Object, NULL, msg->taglist)
			));

		case TA_SmartReq_ExamineFiles:
			return (tr_smartreq_examine(msg->obj,
				(APTR) GetTagData(TT_SmartReq_Data, NULL, msg->taglist),
				(APTR) GetTagData(TT_SmartReq_File1, NULL, msg->taglist),
				(APTR) GetTagData(TT_SmartReq_File2, NULL, msg->taglist)
			));

		case TA_StartupPrefs_Load:
			return (tr_sprefs_load(msg->obj));

		case TA_Print:
			return (tr_print(msg->obj,
				(APTR)GetTagData(TT_Print_Source, NULL, msg->taglist),
				GetTagData(TT_Print_Type, PRT_NONE, msg->taglist)
			));

		default:
			D(PROC, bug("command %ld not supported yet\n", msg->action));
			return (TRUE);
	}
}
#endif
